import os
import time
from selenium import webdriver
import utils
import reader
import constants

class City:
    def __init__(self, driver, name, fast_mode=False):
        self.driver = driver
        self.name = name
        self.save_dir = os.path.join('cities', name)
        if not os.path.exists(self.save_dir):
            os.makedirs(self.save_dir)
        self.fast_mode = fast_mode
        self.center = ''
        self.zoom = ''

    def download_center_map(self):
        self.load_city_map()
        self.center, self.zoom = utils.parse_url(self.driver.current_url)
        if self.center in constants.ZOOM_OUT:
            self.zoom_out()
        self.take_map_screenshot('boundaries.png')

    def zoom_out(self):
        self.zoom = str(int(self.zoom) - 1)
        current_url = self.driver.current_url
        self.driver.get('https://www.google.com/maps/place/{}/@{},{}z/'.format(
        self.name.replace(' ', '+'), self.center.replace(' ', ''), self.zoom))
        while current_url == self.driver.current_url:
            pass

    def load_city_map(self):
        self.driver.get('https://www.google.com/maps/')
        submit_box = utils.wait_until(self.driver.find_element_by_id, 'searchboxinput')
        submit_box.clear()
        submit_box.send_keys(self.name)
        utils.wait_until(self.driver.find_element_by_class_name, 'searchbutton').click()
        while ['https:', '', 'www.google.com', 'maps', 'place'] != self.driver.current_url.split('/')[:5]:
            pass

    def load_map_detailed(self, lat_lng, zoom):
        self.driver.get('https://www.google.com/maps/place/{}/@{},{}z'.format(
        self.name.replace(' ', '+'), lat_lng.replace(' ', ''), zoom))

    def set_corner_information(self):
        self.driver.get('file://{}/js/index.html'.format(os.getcwd()))
        ele = utils.wait_until(self.driver.find_element_by_id, 'cityinfo')
        ele.clear()
        ele.send_keys('{}, {}'.format(self.center, self.zoom))
        self.driver.find_element_by_id('submitbutton').click()
        while not ele.get_attribute('value'):
            pass
        self.boundaries = [float(bound) for bound in ele.get_attribute('value').replace('(', '').replace(')', '').split(', ')]


    def take_map_screenshot(self, filename):
        container = self.driver.find_element_by_id('app-container')
        container.click()
        time.sleep(5)
        self.driver.save_screenshot(os.path.join(self.save_dir, filename))


    def calculate_coordinates(self):
        boundaries_filename = os.path.join(self.save_dir, 'boundaries.png')
        rgb_im = reader.load_image(boundaries_filename)
        nodes = reader.get_nodes(rgb_im, self.fast_mode)
        ordered_nodes = reader.order_nodes(nodes, rgb_im)
        ordered_nodes = utils.prune_nodes(ordered_nodes)
        ordered_nodes = utils.remove_overlapping_polygons(ordered_nodes)
        city_map = reader.Map(self.center, self.boundaries, rgb_im.size[0], rgb_im.size[1])
        city_map.attach_nodes(ordered_nodes)
        reader.write_coordinates(self.save_dir, city_map.nodes)
        with open(os.path.join('js', 'cityarray.js'), 'w') as f:
            f.write('var COORDINATES = [\n')
            for node_collection in ordered_nodes:
                f.write('[\n')
                for node in node_collection:
                    f.write('new google.maps.LatLng({}, {}),\n'.format(node.lat, node.lng))
                f.write('],\n')    
            f.write('];\n')
            f.write('\n')
            f.write('var INFO = "{}, {}";'.format(self.center, self.zoom))


import os
import collections
from PIL import Image
import copy
import requests
import time
import sys
from urllib import request
import utils
import constants

MAX_BLUE_GREEN_DIFFERENCE = 20
VALID = [[232, 151, 142], [237, 162, 155]]
INVALID = [[215, 60, 48], [218, 97, 97], [179, 56, 44], [180, 56, 44]]
# Highway signs (domestic), Hospitals, Highway signs (foreign), city names

def write_coordinates(dir, nodes):
    with open(os.path.join(dir, 'coordinates.txt'), 'w') as f:
        for node_collection in nodes:
            f.write('[\n')
            for node in node_collection:
                f.write('new google.maps.LatLng({}, {}),\n'.format(node.lat, node.lng))
            f.write(']\n')

def load_image(filename):
    im = Image.open(filename)
    im.load()
    return im.convert('RGB')

def get_nodes(rgb_image, fast_mode):
    nodes = []
    width, height = rgb_image.size
    do_not_check = set()
    for x in range(width):
        for y in range(height):
            if ((y in range(constants.MENU_BORDERS[0][0], constants.MENU_BORDERS[0][1] + 1) and
                x in range (constants.MENU_BORDERS[1][0], constants.MENU_BORDERS[1][1] + 1)) or
                (x, y) in do_not_check):
                continue
            r, g, b = rgb_image.getpixel((x, y))
            if [r, g, b] == INVALID[2]:
                exclude_surrounding_nodes(x, y, do_not_check, rgb_image)
            if valid_color(r, g, b) and no_invalid_adjacent(x, y, rgb_image):
                nodes.append(Node(x, y))
                if fast_mode:
                    add_skipped_nodes(x, y, do_not_check)
    nodes[0].visited = True
    return nodes

def exclude_surrounding_nodes(x, y, nodes_to_exclude, rgb_im, depth=5):
    for i in range(-depth, depth + 1):
        for j in range(-depth, depth + 1):
            r, g, b = rgb_im.getpixel((x+i, y+j))
            if [r, g, b] != INVALID[2] or (i == 0 and j == 0):
                try:
                    nodes_to_exclude.add((x+i, y+j))
                except:
                    pass

def valid_color(r, g, b):
    if ([r, g, b] in VALID or
        (r in constants.ALLOWABLE_RED and g in constants.ALLOWABLE_GREEN and
        b in constants.ALLOWABLE_BLUE and abs(g - b) < MAX_BLUE_GREEN_DIFFERENCE)):
        return True
    return False

def no_invalid_adjacent(x, y, image):
    for i in range(-2, 3):
        for j in range(-2, 3):
            try:
                r, g, b = image.getpixel((x + i, y + j))
                if [r, g, b] in INVALID or (r in constants.ALLOWABLE_RED and 100 > g == b):
                    return False
            except IndexError:
                return False
    return True

def order_nodes(nodes, rgb_im):
    ordered_nodes = []
    unvisited = [node for node in nodes if node.visited is False]
    while unvisited:
        node_collection = []
        current = unvisited[0]
        current.visited = True
        closest, distance = get_closest_unvisited_node(current, nodes, rgb_im)
        while closest is not None:
            node_collection.append(current)
            current = closest
            current.visited = True
            closest, distance = get_closest_unvisited_node(current, nodes, rgb_im)
            if closest is None:
                break
            i = -1
            while distance > constants.MAX_NODE_DIFFERENCE:
                if current is node_collection[0] or i < -constants.MAX_NODE_BACKTRACK:
                    closest = None
                    break
                current = node_collection[i]
                closest, distance = get_closest_unvisited_node(current, unvisited, rgb_im)
                i -= 1
        if len(node_collection) > 2:
            ordered_nodes.append(node_collection)
        unvisited = [node for node in nodes if node.visited is False]
    return ordered_nodes

def get_closest_unvisited_node(current, nodes, rgb_im):
    closest_node = None
    shortest_distance = None
    pos = nodes.index(current)
    i = 1
    go_up = True
    go_down = True
    while (0 <= pos - i or len(nodes) > pos + i) and (go_up or go_down):
        signs = []
        if go_down:
            signs.append(-1)
        if go_up:
            signs.append(1)
        for sign in signs:
            index = pos + i*sign
            if not 0 <= index < len(nodes):
                continue
            node = nodes[index]
            if closest_node is not None:
                if sign == -1 and shortest_distance < current.x - node.x:
                    go_down = False
                elif sign == 1 and shortest_distance < node.x - current.x:
                    go_up = False
            if node.visited:
                continue
            distance = utils.get_distance(nodes[pos], node)
            distance *= utils.get_water_multiplier(current, node, rgb_im)
            if shortest_distance is None or distance < shortest_distance:
                closest_node = node
                shortest_distance = distance
        i += 1
    return closest_node, shortest_distance

def add_skipped_nodes(x, y, do_not_check):
    do_not_check.add((x+1, y-1))
    do_not_check.add((x+1, y))
    do_not_check.add((x+1, y+1))
    do_not_check.add((x, y+1))

class Map:
    def __init__(self, center, boundaries, pixwidth, pixheight):
        self.center = center
        self.north = boundaries[2]
        self.east = boundaries[3]
        self.south = boundaries[0]
        self.west = boundaries[1]
        self.pixwidth = pixwidth
        self.pixheight = pixheight
        self.nodes = []
        
    def attach_nodes(self, nodes):
        self.nodes = nodes
        for node_collection in nodes:
            for node in node_collection:
                node.map = self

    @property
    def width(self):
        '''
        Longitudinal width of the map
        '''
        return abs(self.east - self.west)

    @property
    def height(self):
        return abs(self.north - self.south)


class Node:
    def __init__(self, x, y):
        self.x = x
        self.y = y
        self.map = None
        self.visited = False

    @property
    def lat(self):
        return self.map.north - self.y/self.map.pixheight*self.map.height

    @property
    def lng(self):
        return self.map.west + self.x/self.map.pixwidth*self.map.width


    def __str__(self):
        return '<Node at {}, {}>'.format(self.x, self.y)

    def __repr__(self):
        return self.__str__()


if __name__ == '__main__':
    main()


import os
import re
import time
from selenium import webdriver
import reader
import constants

def get_city_list(filename):
    cities = []
    with open(filename) as f:
        for line in f:
            try:
                city_split = line.split(',')
                if city_split[0][-1] == ' ': city_split[0] = city_split[0][:-1]
                cities.append('{}, {}'.format(city_split[0], city_split[1]))
            except IndexError:
                pass
    return cities

def parse_url(url):
    spl = url.split('/')
    assert ['https:', '', 'www.google.com', 'maps', 'place'] == spl[:5]
    lat_lng = re.search(r'@([-0-9.]+,[-0-9.]+)', spl[6]).group(1)
    lat_lng = lat_lng.replace(',', ', ')
    zoom = re.search(r'(\d{1,2})z$', spl[6]).group(1)
    return lat_lng, zoom

def launch_driver(url=''):
    driver = webdriver.Firefox()
    driver.maximize_window()
    if url:
        driver.get(url)
    return driver

def wait_until(func, arg, interval=1):
    while True:
        try:
            return func(arg)
        except:
            time.sleep(interval)

def poly_overlap(poly1, poly2):
    i = -len(poly1) + 1
    for point in poly1:
        mx, my = get_midpoint(point, poly1[i])
        midpoint = reader.Node(mx, my)
        if contains(point, poly2) or contains(midpoint, poly2):
            return True
    return False

def contains(point, polygon):
    assert len(polygon) > 2
    crossings = 0
    i = -len(polygon) + 1
    matched_y = set()
    for node in polygon:
        if ray_crosses_edge(point, node, polygon[i]):
            if not point.y in matched_y:
                matched_y.add(point.y)
                crossings += 1
        i += 1
    return crossings % 2 == 1

def ray_crosses_edge(point, v1, v2):
    # All three args are nodes with x and y properties
    # Ray goes north for this function
    if point.x > max(v1.x, v2.x):
        return False
    if point.x < min(v1.x, v2.x):
        return False
    if point.y < min(v1.y, v2.y):
        return False
    if v1.x == v2.x:
        return False
    slope = (v2.y-v1.y) / (v2.x - v1.x)
    if point.y <= v2.y - (v2.x - point.x) * slope:
        return False
    return True

def remove_overlapping_polygons(polygons):
    assert polygons
    polygons.sort(key=len, reverse=True)
    okay_polygons = [polygons[0]]
    for test_polygon in polygons:
        overlap = False
        if test_polygon in okay_polygons:
            continue
        for okay_polygon in okay_polygons:
            if poly_overlap(test_polygon, okay_polygon):
                overlap = True
                break
        if not overlap:
            okay_polygons.append(test_polygon)
    return okay_polygons

def max_length_list(alist):
    max_list = None
    for sublist in alist:
        if max_list is None or len(sublist) > len(max_list):
            max_list = sublist
    return max_list

def get_water_multiplier(node1, node2, rgb_image):
    '''
    Return approximate water multiplication factor between two nodes
    '''
    x = node1.x
    y = node1.y
    count = 0
    iterations = 10
    for i in range(iterations):
        x += (node2.x - node1.x) / iterations
        y += (node2.y - node1.y) / iterations
        r, g, b = rgb_image.getpixel((x, y))
        if is_water(r, g, b):
            count += 1
    return 1 + count * .5

def is_water(r, g, b):
    if r in constants.WATER_RED and g in constants.WATER_GREEN and b in constants.WATER_BLUE:
        return True
    return False

def prune_nodes(nodes):
    pruned_nodes = []
    for node_collection in nodes:
        assert len(node_collection) > 2
        pruned_collection = node_collection[:2]
        for node in node_collection[2:]:
            if (get_slope(pruned_collection[-2], pruned_collection[-1]) == get_slope(pruned_collection[-2], node) or
                get_distance(pruned_collection[-1], node)) <= 2:
                pruned_collection.pop()
            pruned_collection.append(node)
        if len(pruned_collection) > 2:
            pruned_nodes.append(pruned_collection)
    return pruned_nodes

def get_slope(node1, node2):
    try:
        return (node2.y-node1.y) / (node2.x - node1.x)
    except:
        return None

def get_distance(node1, node2):
    return((node1.x - node2.x)**2 + (node1.y - node2.y)**2)**.5

def get_midpoint(node1, node2):
    return (node1.x + (node2.x - node1.x) / 2, node1.y + (node2.y - node1.y) / 2)


ZOOM_OUT = {
	'40.7590615, -73.969231': 'Manhattan, NY',
}

MAX_NODE_DIFFERENCE = 100
MAX_NODE_BACKTRACK = 50

MENU_BORDERS = [[15, 123], [30, 429]]

WATER_RED = list(range(148, 185))
WATER_GREEN = list(range(193, 215))
WATER_BLUE = list(range(253, 256))

ALLOWABLE_RED = list(range(218, 241))
ALLOWABLE_GREEN = list(range(0, 191))
ALLOWABLE_BLUE = list(range(65, 181))


from setuptools import setup, find_packages # Always prefer setuptools over distutils
from codecs import open # To use a consistent encoding
from os import path
here = path.abspath(path.dirname(__file__))
setup(
	name='gmapsbounds',
	# Versions should comply with PEP440. For a discussion on single-sourcing
	# the version across setup.py and the project code, see
	# http://packaging.python.org/en/latest/tutorial.html#version
	version='0.0.6',
	description='Extract Lat/Lng boundary points of geographical regions from Google Maps',
	# The project's main homepage.
	url='https://github.com/evfredericksen/gmapsbounds',
	# Author details
	author='Evan Fredericksen',
	author_email='evfredericksen@gmail.com',
	# Choose your license
	license='MIT',
	# See https://pypi.python.org/pypi?%3Aaction=list_classifiers
	classifiers=[
	# How mature is this project? Common values are
	# 3 - Alpha
	# 4 - Beta
	# 5 - Production/Stable
	'Development Status :: 3 - Alpha',
	# Indicate who your project is intended for
	'Intended Audience :: Developers',
	'Topic :: Software Development :: Build Tools',
	# Pick your license as you wish (should match "license" above)
	'License :: OSI Approved :: MIT License',
	# Specify the Python versions you support here. In particular, ensure
	# that you indicate whether you support Python 2, Python 3 or both.
	'Programming Language :: Python :: 2',
	'Programming Language :: Python :: 2.6',
	'Programming Language :: Python :: 2.7',
	'Programming Language :: Python :: 3',
	'Programming Language :: Python :: 3.2',
	'Programming Language :: Python :: 3.3',
	'Programming Language :: Python :: 3.4',
	],
	# What does your project relate to?
	keywords='google maps lat lng boundaries',
	# You can just specify the packages manually here if your project is
	# simple. Or you can use find_packages().
	packages=['gmapsbounds'],
	package_dir = {'gmapsbounds': 'gmapsbounds'},
	# List run-time dependencies here. These will be installed by pip when your
	# project is installed. For an analysis of "install_requires" vs pip's
	# requirements files see:
	# https://packaging.python.org/en/latest/technical.html#install-requires-vs-requirements-files
	install_requires=['selenium'],
	# To provide executable scripts, use entry points in preference to the
	# "scripts" keyword. Entry points provide cross-platform support and allow
	# pip to create the appropriate form of executable for the target platform.
	scripts = ['scripts/gmapsbounds.py'],
	# entry_points={
	# 	'console_scripts': [
	# 	'sample=sample:main',
	# 	],
	# },
	 long_description = '''\
	Use Google Maps web app and API to find lat/lng boundaries of
	geographical regions, including cities, ZIP codes, counties,
	states and countries.
	'''
)


import os
import re
import sys
import time
import subprocess
from selenium import webdriver
import selenium.webdriver.support.ui as ui
import obj
import reader
import utils

def main():
    # cities = utils.get_city_list('files/cities.txt')[:3]
    cities = ['Instanbul, Turkey']
    if cities:
        driver = utils.launch_driver()
        try:
            for city_name in cities:
                city = obj.City(driver, city_name)
                city.download_center_map()
                city.center, city.zoom = utils.parse_url(city.driver.current_url)
                city.set_corner_information()
                city.calculate_coordinates()
        finally:
            driver.close()

if __name__ == '__main__':
    main()
